function[finalOutput] = CosinePacketDirectionalSeparate(x,srcDir,D);
% function[finalOutput] =
% CosinePacketDirectionalSeparate(x,srcDir,D);
%
% function to determine directional separation performance of best basis serach of Cosine
% Packets  for comparison with that of  Dual Tree Complex Wavelet Packets 
%
% x: stereo input mixture
% frameSize: size of STFT frame
% D; level of packet decomposition
%
% srcDir: vector of source directions in the range -pi/4 to pi/4, this
% version of this function assumes that they are equidistantly spaced
%
% Jez Wells, jjw100[at]ohm.york.ac.uk
% Please note that this is demonstration code, solely provided to illustrate the
% methods used for the paper A Comparison of Analysis and Synthesis Methods
% for Directional Segmentation of Audio submitted to the DAFx11
% conference.
%
% This function makes use of functions from the Wavelab 850 Toolbox by 	
% David Donoho, Arian Maleki, Morteza Shahram, Jon Buckheit,Maureen Clerc
% Jerome Kalifa, Stephane Mallat and Thomas Yu. It is available at 
% http://www-stat.stanford.edu/~wavelab/

% length of input to CP functions must be dyadic, so extend if neccesary

len = length(x);
p = log2(len);
if(rem(p,1))
    new_length = 2^ceil(p);
    x = [x;zeros(new_length-length(x),2)];
end

leftCP = CPAnalysis(x(:,1),D,'Sine');
rightCP = CPAnalysis(x(:,2),D,'Sine');
clear x
midCP = leftCP+rightCP;
sideCP = rightCP-leftCP;

% compute energies and magnitudes
sz = size(leftCP);
leftEnergy = leftCP.^2;
rightEnergy = rightCP.^2;
totalEnergy = leftEnergy+rightEnergy;
% calculate directions
direction = sign(rightEnergy-leftEnergy).*atan(abs(sideCP./midCP)); 
clear sideCP midCP rightEnergy leftEnergy % don't need these any more
direction(find(isnan(direction))) = 0;

stree = CalcStatTree(totalEnergy,'Entropy');
%stree = directionalCalcStatTree(totalEnergy,direction,srcDir);
clear totalEnergy 
[btree,vtree] = BestBasis(stree,D);
clear stree vtree
pack

% separate sources
[outLeft,outRight,numSources] = hannWindowSeparate(leftCP,rightCP,direction,srcDir);
clear leftCP rightCP

% resynthesize
for n = 1:numSources
    finalOutput{n} = [CPSynthesis(btree,outLeft{n},'Sine');CPSynthesis(btree,outRight{n},'Sine')]';
    % truncate to match length of input
    finalOutput{n} = finalOutput{n}(1:len,:)
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[outLeft,outRight,numSources] = hannWindowSeparate(inLeft,inRight,direction,srcDir);

% uses hann window shape to separate sources, currently the shape of the 
% hann window assumes that sources are equidistantly spaced in the front
% quadrant, this leaves a residual which is in the out of phase region
% (slightly inefficient in that this is done for all packets rather than the
% best basis, but it makes the code easier to read)

numSources = length(srcDir);
outLeft = cell(numSources,1);
outRight = cell(numSources,1);

for n = 1:numSources
    % allocate memory
    outLeft{n} = zeros(size(inLeft));
    outRight{n} = zeros(size(inRight));
    % apply hann window centred on source direction 
    % first ensure that -pi/2 wraps round into pi/2
	direction(find(direction-srcDir(n)>=pi/2))...
                    = direction(find(direction-srcDir(n)>=pi/2))-pi;
	direction(find(direction-srcDir(n)<-pi/2))...
                    = direction(find(direction-srcDir(n)<-pi/2))+pi;
	% don't want window to be periodic (i.e. should be zero
	% outside of central lobe) so do this only for indices that
	% are close to srcDir
    idx = find(abs(direction-srcDir(n))<pi/(numSources*2)); 
    multiplier = 0.5*(1+cos(numSources*2*(direction(idx)-srcDir(n))));
    outLeft{n}(idx) = inLeft(idx).*multiplier;  
    outLeft{n} = sparse(outLeft{n});    % to prevent 'out of memory' errors
    outRight{n}(idx) = inRight(idx).*multiplier;
    outRight{n} = sparse(outRight{n});
end

% clear some memory and then 'de-sparsify' matrices
% as part of 'out of memory' avoidance
clear inLeft inRight
for n = 1:numSources
    outLeft{n} = full(outLeft{n});
    outRight{n} = full(outRight{n});
end
    





