function[finalOutput] = STFTDirectionalSeparate(x,srcDir,frameSize,overlap,OLAmethod);
% function [finalOutput] =
% STFTDirectionalSeparate(x,srcDir,frameSize,overlap);
%
% function to determine directional separation performance of the STFT
% for comparison with that of  Dual Tree Complex Wavelet Packets 
%
% x: stereo input mixture
% frameSize: size of STFT frame
% overlap: extent of overlap between adjacent frames (needs to be a minimum
% of 2 for the Hann window)
% srcDir: vector of source directions in the range -pi/4 to pi/4, this
% version of this function assumes that they are equidistantly spaced
% OLAmethod: Determines whether synthesized frames are tapered before being
% combined by overlap-add (OLA). 'taper': tapering is applied before OLA,
% 'no_taper': no tapering is applied before OLA
%
% Jez Wells, jjw100[at]ohm.york.ac.uk
% Please note that this is demonstration code, solely provided to illustrate the
% methods used for the paper A Comparison of Analysis and Synthesis Methods
% for Directional Segmentation of Audio submitted to the DAFx11
% conference.


if(min(size(x))~=2)
    error('input must be two channel');
end
if(rem(frameSize,2))
    error('frameSize must be an even number');
end
if(nargin<5)
    if(overlap>=4)
        % taper output frame after iFFT
        OLAmethod = 'taper';
    else
         % do not taper if there is insufficient overlap
         OLAmethod = 'no_taper';
    end
end

% generate window
Hann = hann(frameSize);

frameCount = 1;
samples = length(x);
complexSize = (frameSize/2)+1;

% ensure the output matrix will be big enough to accomodate a whole number of
% frames by rounding up
outSamples = (ceil(samples/frameSize) + 1) * frameSize;
finalOutput = cell(length(srcDir),1);
for n = 1:length(srcDir)
    finalOutput{n} = zeros(outSamples,2);
end

while frameCount < samples
    outPointer = frameCount + (frameSize-1);
    if outPointer > samples
        outPointer = samples;
        siz = outPointer - frameCount + 1;
    end
    frame = x(frameCount:outPointer,:);
  
    if length(frame) < frameSize
        %calculate difference between input and frame length
        diff = frameSize - length(frame);
        %append zeros
        oo = zeros(diff,2);
        frame = [rot90(frame), rot90(oo)];
        %rotate back after appending zeros
        frame = rot90(frame,-1);
    end
    %place individual channels in separate vectors and apply window
    left = frame(:,1).*Hann;  
    right = frame(:,2).*Hann;
    
    % perform fft
    leftFFT = fft(left,frameSize);
    rightFFT = fft(right,frameSize);
    % extract magnitudes, energies and phases for each bin
    leftMag = abs(leftFFT);
    leftEnergy = leftMag.^2;
    rightMag = abs(rightFFT);
    rightEnergy = rightMag.^2;
    midMag = abs(leftFFT+rightFFT);
    sideMag = abs(rightFFT-leftFFT);
    
    % calculate directions for each bin
    direction = sign(rightEnergy-leftEnergy).*atan(sideMag./midMag); 
    direction(find(isnan(direction))) = 0;
   
    [leftFFTout,rightFFTout,numSources] = ...
    hannWindowSeparate(leftFFT,rightFFT,direction,srcDir,frameSize);
    
    % now convert these directional separated FFTs back to the time domain
    % by overlap-add
    for n = 1:numSources
        leftOutput = ifft(leftFFTout{n}, 'symmetric');    %inverse FFT of data with complex conjugate symmetry is faster
        rightOutput = ifft(rightFFTout{n}, 'symmetric');
        if(strcmp(OLAmethod,'taper'))
            % taper output
            leftOutput = leftOutput.*Hann;
            rightOutput = rightOutput.*Hann;
        end
        output(:,1) = leftOutput;
        output(:,2) = rightOutput;
        %place finalOutput in single matrix to be written to stereo .wav file    
        finalOutput{n}(frameCount:(frameCount + (frameSize-1)),:) =  finalOutput{n}(frameCount:(frameCount + (frameSize-1)),:) + output(1:(frameSize),:);
    end
    frameCount = frameCount + floor(frameSize/overlap);
end

% truncate output so that it is same length as input
for n = 1:numSources
    finalOutput{n}(1:samples,:);
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[leftFFTout,rightFFTout,numSources] = hannWindowSeparate(leftFFTin,...
    rightFFTin,direction,srcDir,N);

% uses hann window shape to separate sources, currently the shape of the 
% hann window assumes that sources are equidistantly spaced in the front
% quadrant, this leaves a residual which is in the out of phase region

numSources = length(srcDir);
leftFFTout = cell(numSources,1);
rightFFTout = cell(numSources,1);

for n = 1:numSources
    % each different position into the -pi/2 to pi/2 interval
    leftFFTout{n} = zeros(N,1);
    rightFFTout{n} = zeros(N,1);
    % apply hann window centred on source direction 
    % first ensure that -pi/2 wraps round into pi/2
    direction(find(direction-srcDir(n)>=pi/2))...
    = direction(find(direction-srcDir(n)>=pi/2))-pi;
    direction(find(direction-srcDir(n)<-pi/2))...
    = direction(find(direction-srcDir(n)<-pi/2))+pi;
    % don't want window to be periodic (i.e. should be zero
    % outside of central lobe) so do this only for indices that
    % are close to srcDir
    idx = find(abs(direction-srcDir(n))<pi/(numSources*2)); 
    multiplier = 0.5*(1+cos(numSources*2*(direction(idx)-srcDir(n))));
    leftFFTout{n}(idx) = leftFFTin(idx).*multiplier;
    rightFFTout{n}(idx) = rightFFTin(idx).*multiplier;
end

