function [output] = complexWaveletsDirectionalSeparate(x,srcDir,type,N,filters);
% function [output] =
% complexWaveletsDirectionalSeparate(x,srcDir,type,N,filters);
%
% function to determine directional separation performance of the STFT
% for comparison with that of  Dual Tree Complex Wavelet Packets 
%
% x: stereo input mixture
%
% N: maximum decomposition level (number of scales)
%
% type: 'wpd' gives wavelet packet decomposition (WPD) with best basis 
% search  using standard (Coifman and Wickerhauser) entropy measure
% 'phase' gives WPD with best basis search via phase-weighted entropy
% measure
% 'DWT' gives standard DT-CDWT (fixed dyadic decomposition)
% 
% srcDir: vector of source directions in the range -pi/4 to pi/4, this
% version of this function assumes that they are equidistantly spaced
%
% filters: there are two sets of filters to choose from: 'short' and 'long'
% These are the filters provided at: %
% (http://www2.itu.edu.tr/~ibayram/dtcwpt/)
%
% Jez Wells, jjw100[at]ohm.york.ac.uk
% Please note that this is demonstration code, solely provided to illustrate the
% methods used for the paper A Comparison of Analysis and Synthesis Methods
% for Directional Segmentation of Audio submitted to the DAFx11
% conference.
%
% This function makes use of functions from the Dual-Tree Complex Wavelet 
% Packet Toolbox by I. Bayram which is available at:
% http://www2.itu.edu.tr/~ibayram/dtcwpt/
% One function from that toolbox has been adapted for this work and is 
% included with the files made available for this paper: IDTWPT_JJW.m

% divide by zero will happen for any purely out of phase stuff
warning off MATLAB:divideByZero

if(nargin<5)
    filters = 'set1';
end
if(nargin<4)
    N = 11; % decomposition level
end
if(nargin<5)
    type = 'WPD';
end

% load wavelet filters
if(strcmp(filters, 'set1'))
    load dtcwpt_filters; % (http://www2.itu.edu.tr/~ibayram/dtcwpt/)
elseif(strcmp(filters, 'set2'))
    load dtcwpt_filters_long; % (http://www2.itu.edu.tr/~ibayram/dtcwpt/)
elseif(strcmp(filters, 'set3'))
    load dtcwpt_filters_24;  
end

% append zeros to ensure linear convolution on reconstruction
lenX = length(x);
temp = zeros(length(x)+((2^(N+1))*length(g)),2);
temp(1:lenX,:) = x;
x = temp;
clear temp

% compute the four wavelet transforms (left, right, real and
% imaginary)
yleft_real = DTWPT(x(:,1)',first_1,h,f,N);
yleft_imag = DTWPT(x(:,1)',first_2,g,f,N);
yright_real = DTWPT(x(:,2)',first_1,h,f,N);
yright_imag = DTWPT(x(:,2)',first_2,g,f,N);
% compute energies and magnitudes
sz = size(yleft_real);
yleft_en = cell(sz);
yright_en = cell(sz);
ymid_mag = cell(sz);
yside_mag = cell(sz);
ytotal_en = cell(sz);
len = sz(1)*sz(2);
for n = 1:len
    yleft_en{n} = yleft_real{n}.^2+yleft_imag{n}.^2;
    yright_en{n} = yright_real{n}.^2+yright_imag{n}.^2;
    ytotal_en{n} = yleft_en{n}+yright_en{n};
    ymid_mag{n} = sqrt((yleft_real{n}+yright_real{n}).^2+(yleft_imag{n}+yright_imag{n}).^2);
    yside_mag{n} = sqrt((yright_real{n}-yleft_real{n}).^2+(yright_imag{n}-yleft_imag{n}).^2);
end
% calculate directions and absolute phase difference
phi_diff = cell(sz);
direction = cell(sz);
for n = 1:len
    direction{n} = zeros(size(ymid_mag{n}));
    direction{n}(:) = sign(yright_en{n}(:)-yleft_en{n}(:)).*atan(yside_mag{n}(:)./ymid_mag{n}(:)); 
    direction{n}(find(isnan(direction{n}))) = 0;
    % calculate absolute phase difference
    leftPhi = atan2(yleft_imag{n},yleft_real{n});
    rightPhi = atan2(yright_imag{n},yright_real{n});
    phi_diff{n} = abs(abs(rightPhi-leftPhi)-pi);  % wrap differences of > pi back in
    phi_diff{n}(find(phi_diff{n}==0)) = eps;  % remove zeros
end

if(strcmp(type,'WPD'))
    best_basis = entropyBasisCalc(sz,len,N,ytotal_en);
elseif(strcmp(type,'phase'))
    best_basis = phaseWeightedEntropyBasisCalc(sz,len,N,ytotal_en,phi_diff);
elseif(strcmp(type,'DWT'))
    best_basis = fixStandardDyadicBasis(sz,N);
end

% now search for the upmost marked node, and set all generations below it
% to zero
for n = 1:N
    for nn = 1:2^n
        if(best_basis(n,nn))
            % set all generations below to zero
            for n3 = n+1:N
                best_basis(n3,((2^(n3-n))*(nn-1))+1:((2^(n3-n))*nn)) = 0;
            end
        end
    end
end

% separate sources
[outLeftReal,outLeftImag,outRightReal,outRightImag,numSources] = ...
    hannWindowSeparate(yleft_real,yleft_imag,yright_real,yright_imag,...
    direction,srcDir,best_basis,N);

% now create 'cumulative' mask to display tree
for n = N:-1:2
    for nn = 1:2:2^n
        if(best_basis(n,nn))
            % set all generations above to one
            best_basis(n-1,ceil(nn/2)) = 1;
        end
    end
end


% resynthesize
for n = 1:numSources
    output{n}(:,1) = 0.5*(IDTWPT_JJW(outLeftReal{n},first_1(:,end:-1:1),h(:,end:-1:1),f(:,end:-1:1),best_basis)...
        + IDTWPT_JJW(outLeftImag{n},first_2(:,end:-1:1),g(:,end:-1:1),f(:,end:-1:1),best_basis));
    output{n}(:,2) = 0.5*(IDTWPT_JJW(outRightReal{n},first_1(:,end:-1:1),h(:,end:-1:1),f(:,end:-1:1),best_basis)...
        + IDTWPT_JJW(outRightImag{n},first_2(:,end:-1:1),g(:,end:-1:1),f(:,end:-1:1),best_basis));
    % truncate to same length as input
    output{n} = output{n}(1:lenX,:);
end

displayTree(best_basis,N);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[best_basis] = fixStandardDyadicBasis(sz,N);
% function that returns the 'standard' dyadic wavelet basis
best_basis = zeros(sz);
for n = 1:N
    % standard decomposition uses detail (HPF output) at each scale
    best_basis(n,2) = 1;
end
% highest scale uses both detail (HPF) and approximation (LPF)
best_basis(n,1) = 1;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[] = displayTree(best_basis,N);
disp = ones(size(best_basis));
for n = 1:N
    for nn = 1:2^n
        disp(n,(1+((nn-1)*(2^(N-n)))):nn*(2^(N-n))) = best_basis(n,nn);
    end
end
imagesc(linspace(0,22050,2048),1:N,disp)
colormap(flipud(gray))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[best_basis] = entropyBasisCalc(sz,len,N,ytotal_en);
warning off MATLAB:divideByZero

entropy = zeros(sz);

for n = 1:len
    % compute the shannon entropy
    entropy(n) = -sum(ytotal_en{n}.*log(eps+ytotal_en{n}));
end

% now compare entropy of children with those of parents
% follows outline of algorithm in Wickerhauser p.284-285
best_basis = zeros(sz);
% start by marking all bottom nodes
best_basis(N,:) = 1;
for n = N:-1:2
    for nn = 1:2^(n-1)
        % entropy is additive, so entropy of kids is simply the summation
        % of their entropies
        sumEntropy = entropy(n,(2*nn)-1)+entropy(n,2*nn);
        if(sumEntropy<entropy(n-1,nn)) % looking for lowest entropy value
            % move new entropy values to parent
            entropy(n-1,nn) = sumEntropy;
        else
            % parent has better kurtosis than combined kids so mark parent
            best_basis(n-1,nn) = 1;  
            % set kids to zero
            best_basis(n,(2*nn)-1:2*nn) = 0;
        end
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[best_basis] = phaseWeightedEntropyBasisCalc(sz,len,N,ytotal_en,phi_diff);
warning off MATLAB:divideByZero

entropy = zeros(sz);

for n = 1:len
    % compute the phase-weighted shannon entropy
    entropy(n) = -sum(ytotal_en{n}.*log(eps+ytotal_en{n})./(phi_diff{n}+0.01));
end

% now compare entropy of children with those of parents
% follows outline of algorithm in Wickerhauser p.284-285
best_basis = zeros(sz);
% start by marking all bottom nodes
best_basis(N,:) = 1;
for n = N:-1:2
    for nn = 1:2^(n-1)
        % entropy is additive, so entropy of kids combined is simply the summation
        % of their entropies
        sumEntropy = entropy(n,(2*nn)-1)+entropy(n,2*nn);
        if(sumEntropy<entropy(n-1,nn)) % looking for lowest entropy value
            % move new entropy values to parent
            entropy(n-1,nn) = sumEntropy;
        else
            % parent has better entropy than combined kids so mark parent
            best_basis(n-1,nn) = 1;  
            % set kids to zero
            best_basis(n,(2*nn)-1:2*nn) = 0;
        end
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function[outLeftReal,outLeftImag,outRightReal,outRightImag,numSources] = ...
    hannWindowSeparate(yleft_real,yleft_imag,yright_real,yright_imag,...
    direction,srcDir,best_basis,N);

% uses hann window shape to separate sources, currently the shape of the 
% hann window assumes that sources are equidistantly spaced in the front
% quadrant, this leaves a residual which is in the out of phase region

numSources = length(srcDir);
outLeftReal = cell(numSources,1);
outLeftImag = cell(numSources,1);
outRightReal = cell(numSources,1);
outRightImag = cell(numSources,1);

for n = 1:numSources
    % each different position into the -pi/2 to pi/2 interval
    for nn = 1:N
        for n3 = 1:2^nn
            coeffSize = size(yleft_real{nn,n3});
            outLeftReal{n}{nn,n3} = zeros(coeffSize);
            outLeftImag{n}{nn,n3} = zeros(coeffSize);
            outRightReal{n}{nn,n3} = zeros(coeffSize);
            outRightImag{n}{nn,n3} = zeros(coeffSize);
            if(best_basis(nn,n3))
                % apply hann window centred on source direction 
                % first ensure that -pi/2 wraps round into pi/2
                direction{nn,n3}(find(direction{nn,n3}-srcDir(n)>=pi/2))...
                    = direction{nn,n3}(find(direction{nn,n3}-srcDir(n)>=pi/2))-pi;
                direction{nn,n3}(find(direction{nn,n3}-srcDir(n)<-pi/2))...
                    = direction{nn,n3}(find(direction{nn,n3}-srcDir(n)<-pi/2))+pi;
                % don't want window to be periodic (i.e. should be zero
                % outside of central lobe) so do this only for indices that
                % are close to srcDir
                idx = find(abs(direction{nn,n3}-srcDir(n))<pi/(numSources*2)); 
                multiplier = 0.5*(1+cos(numSources*2*(direction{nn,n3}(idx)-srcDir(n))));
                outLeftReal{n}{nn,n3}(idx) = yleft_real{nn,n3}(idx).*multiplier;
                outLeftImag{n}{nn,n3}(idx) = yleft_imag{nn,n3}(idx).*multiplier;
                outRightReal{n}{nn,n3}(idx) = yright_real{nn,n3}(idx).*multiplier;
                outRightImag{n}{nn,n3}(idx) = yright_imag{nn,n3}(idx).*multiplier;
            end
        end
    end 
end





        
    
    